local LubStub = LibStub
local DeathClock = LibStub('AceAddon-3.0'):NewAddon('DeathClock', 'AceEvent-3.0', 'AceConsole-3.0')
_G.DeathClock = DeathClock
local SML = LibStub("LibSharedMedia-3.0")
local debug = false
local blizTargetParent = TargetFrameHealthBar
local blizFocusParent = FocusFrameHealthBar
local blizBoss1Parent = Boss1TargetFrameHealthBar
local blizBoss2Parent = Boss2TargetFrameHealthBar
local blizBoss3Parent = Boss3TargetFrameHealthBar
local blizBoss4Parent = Boss4TargetFrameHealthBar

local function GetTableDefaults()
	return { 
		show = true,
		background = {r = 0,b = 0,g = 0},
		foreground = {r = 1,b = 1,g = 1},
		backgroundAlpha = .25,
		foregroundAlpha = 1,
		font = "DorisPP",
		size = 10,
		outline = 'OUTLINE',
		height = 16,
		width = 51,
		x = -50,
		y = -12,
		p1 = "LEFT",
		p2 = "LEFT",
	}
end

local function GetDefaults()
	return {
		global = {
			targetFrame = GetTableDefaults(),
			focusFrame = GetTableDefaults(),
			boss1Frame = GetTableDefaults(),
			boss2Frame = GetTableDefaults(),
			boss3Frame = GetTableDefaults(),
			boss4Frame = GetTableDefaults()
		}
	}
end

local timetodieUpdate = function(frame, e, unit) 
	local health = UnitHealth(frame.target)
	local time = GetTime()

	if health ~= 0 and frame:GetParent():IsVisible() ~= 1 then
		frame:SetParent(UIParent)
	end
	
	if health ~= 0 and frame:IsShown() ~= 1 and frame.configShow == true then
		frame:Show()
	elseif health == 0 and frame:IsShown() == 1 then
		frame:Hide()
	end
	
	if e == frame.resetEvent or health == UnitHealthMax(frame.target) then
		frame.health0, frame.time0, frame.mhealth, frame.mtime = nil
		frame.text:SetText("...")
		return
	end

	if not frame.health0 then
		frame.health0, frame.time0 = health, time
		frame.mhealth, frame.mtime = health, time
		return
	end

	frame.mhealth = (frame.mhealth + health) * .5
	frame.mtime = (frame.mtime + time) * .5

	if frame.mhealth >= frame.health0 then
		frame.text:SetText("...")
		frame.health0, frame.time0, frame.mhealth, frame.mtime = nil
	else
		time = ceil(health * (frame.time0 - frame.mtime) / (frame.mhealth - frame.health0) )
		if time < 60 then
			frame.text:SetText(format('%ds',time))
		elseif time < 3600 then
			frame.text:SetText(format('%dm%02ds', 1/60 * time, time % 60))
		elseif time < 86400 then
			frame.text:SetText(format('%dh%02dm', 1/3600 * time, 1/60 * (time % 3600)))
		else
			frame.text:SetText(format('%dd%02dh', 1/86400 * time, 1/3600 * (time % 86400)))
		end
	end
	
	if debug == true then
		print(frame:GetName(), frame.text:GetText())
	end
end

function DeathClock:SetupFrame(frameName,parent,textName,target,settingsKey,resetEvent)
	local frame = CreateFrame("Frame", frameName, parent)
	frame:SetPoint("LEFT", parent, "LEFT", -50, -12)
	frame:SetWidth(51)
	frame:SetHeight(16)
	frame:SetMinResize(50,15)
	frame:SetMaxResize(300,90)
	frame:SetResizable(true)
	frame:SetFrameStrata("HIGH")
	frame.texture = frame:CreateTexture("bg", "BACKGROUND", "ActionBarButtonTemplate")
	frame.texture:SetAllPoints(frame)
	frame.texture:SetTexture(0,0,0,.8)

	frame.text = frame:CreateFontString(textName, "OVERLAY")
	frame.text:SetAllPoints(frame)
	frame.text:SetFontObject(TextStatusBarText)
	frame.text:SetJustifyH("RIGHT")
	if resetEvent then
		frame:RegisterEvent(resetEvent)
		frame.resetEvent = resetEvent
	else
		frame.resetEvent = "NoResetEvent"
	end
	frame:SetScript("OnShow", 
		function(frame) 
			frame.health0, frame.time0, frame.mhealth, frame.mtime = nil
			frame:RegisterEvent("UNIT_HEALTH") 
		end
	)
	frame:SetScript("OnHide", 
		function(frame) 
			frame:UnregisterEvent("UNIT_HEALTH")
		end
	)
	frame:EnableMouse(true)
	frame:RegisterForDrag("LeftButton")
	frame:SetClampedToScreen(true)
	frame:SetMovable(true)
	frame:SetScript("OnDragStart", 
		function(frame) 
			if IsAltKeyDown() then 
				frame:StartMoving() 
			elseif IsShiftKeyDown() then
				frame:StartSizing()
			end 
		end
	)
	frame:SetScript("OnDragStop", 
		function(frame) 
			frame:StopMovingOrSizing() 
			local settings = DeathClock.db.global[frame.settingsKey]
			settings.width = frame:GetWidth()
			settings.height = frame:GetHeight()
			local p1,_,p2,x,y = frame:GetPoint()
			settings.x = x
			settings.y = y
			settings.p1 = p1
			settings.p2 = p2
		end
	)
	frame:SetScript("OnEvent", timetodieUpdate)
	frame.target = target
	frame.settingsKey = settingsKey
	return frame
end

function DeathClock:ApplyFrameSettings(frame)
	local settings = self.db.global[frame.settingsKey]
	frame:ClearAllPoints()
	frame:SetWidth(settings.width)
	frame:SetHeight(settings.height)
	frame:SetPoint(settings.p1, frame:GetParent(), settings.p2, settings.x, settings.y)
	frame.text:SetFont(SML:Fetch('font', settings.font), settings.size, settings.outline)
	frame.text:SetTextColor(settings.foreground.r, settings.foreground.g, settings.foreground.b, settings.foregroundAlpha)
	frame.texture:SetTexture(settings.background.r, settings.background.g, settings.background.b, settings.backgroundAlpha)
	frame.configShow = settings.show
	if settings.show == true then
		frame:Show()
	else
		frame:Hide()
	end
end

function DeathClock:OnInitialize()
	self.db = LibStub('AceDB-3.0'):New('DeathClockDB', GetDefaults(), 'Default')
	self:RegisterOptions()
	
	self.t_dc =  self:SetupFrame("TargetDeathClock",blizTargetParent,"TargetDeathClockText","target","targetFrame","PLAYER_TARGET_CHANGED")
	self.f_dc =  self:SetupFrame("FocusDeathClock", blizFocusParent, "FocusDeathClockText", "focus", "focusFrame", "PLAYER_FOCUS_CHANGED")
	self.b1_dc = self:SetupFrame("Boss1DeathClock", blizBoss1Parent, "Boss1DeathClockText", "boss1", "boss1Frame", nil)
	self.b2_dc = self:SetupFrame("Boss2DeathClock", blizBoss2Parent, "Boss2DeathClockText", "boss2", "boss2Frame", nil)
	self.b3_dc = self:SetupFrame("Boss3DeathClock", blizBoss3Parent, "Boss3DeathClockText", "boss3", "boss3Frame", nil)
	self.b4_dc = self:SetupFrame("Boss4DeathClock", blizBoss4Parent, "Boss4DeathClockText", "boss4", "boss4Frame", nil)
	
	self:ApplySettings()
	self:RegisterChatCommand("dc", "SlashCommand")
	self:RegisterChatCommand("deathclock", "SlashCommand")
end

function DeathClock:Debug()
	if debug == true then
		debug = false
	else
		debug = true
	end
end

local function GetFontList()
	return AceGUIWidgetLSMlists.font
end

local function GetValue(info)
	return info.handler.db.global[info.arg[1]][info.arg[2]]
end

local function SetValue(info, value)
	local self = info.handler
	local global = self.db.global
	global[info.arg[1]][info.arg[2]] = value
	self:ApplySettings()
end

local function GetColorValue(info)
	local color = info.handler.db.global[info.arg[1]][info.arg[2]]
	return color.r, color.g, color.b
end

local function SetColorValue(info, r, g, b)
	local self = info.handler
	local global = self.db.global
	global[info.arg[1]][info.arg[2]].r = r
	global[info.arg[1]][info.arg[2]].g = g
	global[info.arg[1]][info.arg[2]].b = b
	self:ApplySettings()
end

local function RestoreFrameDefaults(frame)
	DeathClock.db.global[frame] = GetTableDefaults()
	DeathClock:ApplySettings()
end

local function RestoreDefaults(info)
	DeathClock.db.global[info.arg] = GetTableDefaults()
	DeathClock:ApplySettings()
end

function DeathClock:ApplySettings()
	self:ApplyFrameSettings( self.t_dc)
	self:ApplyFrameSettings( self.f_dc)
	self:ApplyFrameSettings( self.b1_dc)
	self:ApplyFrameSettings( self.b2_dc)
	self:ApplyFrameSettings( self.b3_dc)
	self:ApplyFrameSettings( self.b4_dc)
end

local options = {
	name = 'Death Clock',
	type = 'group',
	desc = 'Display an estimated kill time for your target, focus target and any mobs in the blizzard boss frames.',
	set = SetValue,
	get = GetValue,
	handler = DeathClock,
	args = {
		["TARGET_FRAME"] = {
			name = "Target",
			type = 'group',
			order = 100,
			guiInline = true,
			set = SetValue,
			get = GetValue,
			args = {
				["Show"] = {
					name = "Show",
					type = 'toggle',
					order = 100,
					arg = {"targetFrame",'show'},
					disabled = false,
				},
				["Font Size"] = {
					name = "Font Size",
					type = 'range',
					min = 4, max = 20, step = 1,
					order = 150,
					arg = {"targetFrame",'size'},
				},
				["Font"] = {
					name = "Font",
					type = 'select',
					dialogControl = 'LSM30_Font',
					values = GetFontList,
					order = 200,
					arg = {"targetFrame",'font'},
				},
				["FColor"] = {
					name = "Text Color",
					type = 'color',
					set = SetColorValue,
					get = GetColorValue,
					order = 350,
					arg = {"targetFrame",'foreground'},
				},
				["TextAlpha"] = {
					name = "Text Alpha",
					type = 'range',
					min = 0, max = 1, step = .05,
					order = 400,
					arg = {"targetFrame",'foregroundAlpha'},
				},
				["Outline"] = {
					name = "Outline",
					type = 'select',
					desc = "Set font outline.",
					values = {[''] = NONE, OUTLINE = VOICE_CHAT_NORMAL, THICKOUTLINE = "Thick"},
					order = 425,
					arg = {"targetFrame",'outline'},
				},
				["BColor"] = {
					name = "Background",
					type = 'color',
					set = SetColorValue,
					get = GetColorValue,
					order = 450,
					arg = {"targetFrame",'background'},
				},
				["BAlpha"] = {
					name = "Background Alpha",
					type = 'range',
					min = 0, max = 1, step = .05,
					order = 500,
					arg = {"targetFrame",'backgroundAlpha'},
				},
				["Reset"] = {
					name = "Restore Defaults",
					func = RestoreDefaults,
					type = 'execute',
					order = 550,
					arg = "targetFrame",
				},
			}
		},
	
		["FOCUS_FRAME"] = {
			name = "Focus",
			type = 'group',
			order = 150,
			guiInline = true,
			set = SetValue,
			get = GetValue,
			args = {
				["Show"] = {
					name = "Show",
					type = 'toggle',
					order = 100,
					arg = {"focusFrame",'show'},
					disabled = false,
				},
				["Font Size"] = {
					name = "Font Size",
					type = 'range',
					min = 4, max = 20, step = 1,
					order = 150,
					arg = {"focusFrame",'size'},
				},
				["Font"] = {
					name = "Font",
					type = 'select',
					dialogControl = 'LSM30_Font',
					values = GetFontList,
					order = 200,
					arg = {"focusFrame",'font'},
				},
				["FColor"] = {
					name = "Text Color",
					type = 'color',
					set = SetColorValue,
					get = GetColorValue,
					order = 350,
					arg = {"focusFrame",'foreground'},
				},
				["TextAlpha"] = {
					name = "Text Alpha",
					type = 'range',
					min = 0, max = 1, step = .05,
					order = 400,
					arg = {"focusFrame",'foregroundAlpha'},
				},
				["Outline"] = {
					name = "Outline",
					type = 'select',
					desc = "Set font outline.",
					values = {[''] = NONE, OUTLINE = VOICE_CHAT_NORMAL, THICKOUTLINE = "Thick"},
					order = 425,
					arg = {"focusFrame",'outline'},
				},
				["BColor"] = {
					name = "Background",
					type = 'color',
					set = SetColorValue,
					get = GetColorValue,
					order = 450,
					arg = {"focusFrame",'background'},
				},
				["BAlpha"] = {
					name = "Background Alpha",
					type = 'range',
					min = 0, max = 1, step = .05,
					order = 500,
					arg = {"focusFrame",'backgroundAlpha'},
				},
				["Reset"] = {
					name = "Restore Defaults",
					func = RestoreDefaults,
					type = 'execute',
					order = 550,
					arg = "focusFrame",
				},
			}
		},	
		
		["BOSS_FRAME1"] = {
			name = "Boss 1",
			type = 'group',
			order = 200,
			guiInline = true,
			set = SetValue,
			get = GetValue,
			args = {
				["Show"] = {
					name = "Show",
					type = 'toggle',
					order = 100,
					arg = {"boss1Frame",'show'},
					disabled = false,
				},
				["Font Size"] = {
					name = "Font Size",
					type = 'range',
					min = 4, max = 20, step = 1,
					order = 150,
					arg = {"boss1Frame",'size'},
				},
				["Font"] = {
					name = "Font",
					type = 'select',
					dialogControl = 'LSM30_Font',
					values = GetFontList,
					order = 200,
					arg = {"boss1Frame",'font'},
				},
				["FColor"] = {
					name = "Text Color",
					type = 'color',
					set = SetColorValue,
					get = GetColorValue,
					order = 350,
					arg = {"boss1Frame",'foreground'},
				},
				["TextAlpha"] = {
					name = "Text Alpha",
					type = 'range',
					min = 0, max = 1, step = .05,
					order = 400,
					arg = {"boss1Frame",'foregroundAlpha'},
				},
				["Outline"] = {
					name = "Outline",
					type = 'select',
					desc = "Set font outline.",
					values = {[''] = NONE, OUTLINE = VOICE_CHAT_NORMAL, THICKOUTLINE = "Thick"},
					order = 425,
					arg = {"boss1Frame",'outline'},
				},
				["BColor"] = {
					name = "Background",
					type = 'color',
					set = SetColorValue,
					get = GetColorValue,
					order = 450,
					arg = {"boss1Frame",'background'},
				},
				["BAlpha"] = {
					name = "Background Alpha",
					type = 'range',
					min = 0, max = 1, step = .05,
					order = 500,
					arg = {"boss1Frame",'backgroundAlpha'},
				},
				["Reset"] = {
					name = "Restore Defaults",
					func = RestoreDefaults,
					type = 'execute',
					order = 550,
					arg = "boss1Frame",
				},
			}
		},
		
		
		["BOSS_FRAME2"] = {
			name = "Boss 2",
			type = 'group',
			order = 250,
			guiInline = true,
			set = SetValue,
			get = GetValue,
			args = {
				["Show"] = {
					name = "Show",
					type = 'toggle',
					order = 100,
					arg = {"boss2Frame",'show'},
					disabled = false,
				},
				["Font Size"] = {
					name = "Font Size",
					type = 'range',
					min = 4, max = 20, step = 1,
					order = 150,
					arg = {"boss2Frame",'size'},
				},
				["Font"] = {
					name = "Font",
					type = 'select',
					dialogControl = 'LSM30_Font',
					values = GetFontList,
					order = 200,
					arg = {"boss2Frame",'font'},
				},
				["FColor"] = {
					name = "Text Color",
					type = 'color',
					set = SetColorValue,
					get = GetColorValue,
					order = 350,
					arg = {"boss2Frame",'foreground'},
				},
				["TextAlpha"] = {
					name = "Text Alpha",
					type = 'range',
					min = 0, max = 1, step = .05,
					order = 400,
					arg = {"boss2Frame",'foregroundAlpha'},
				},
				["Outline"] = {
					name = "Outline",
					type = 'select',
					desc = "Set font outline.",
					values = {[''] = NONE, OUTLINE = VOICE_CHAT_NORMAL, THICKOUTLINE = "Thick"},
					order = 425,
					arg = {"boss2Frame",'outline'},
				},
				["BColor"] = {
					name = "Background",
					type = 'color',
					set = SetColorValue,
					get = GetColorValue,
					order = 450,
					arg = {"boss2Frame",'background'},
				},
				["BAlpha"] = {
					name = "Background Alpha",
					type = 'range',
					min = 0, max = 1, step = .05,
					order = 500,
					arg = {"boss2Frame",'backgroundAlpha'},
				},
				["Reset"] = {
					name = "Restore Defaults",
					func = RestoreDefaults,
					type = 'execute',
					order = 550,
					arg = "boss2Frame",
				},
			}
		},	
		
		
		["BOSS_FRAME3"] = {
			name = "Boss 3",
			type = 'group',
			order = 300,
			guiInline = true,
			set = SetValue,
			get = GetValue,
			args = {
				["Show"] = {
					name = "Show",
					type = 'toggle',
					order = 100,
					arg = {"boss3Frame",'show'},
					disabled = false,
				},
				["Font Size"] = {
					name = "Font Size",
					type = 'range',
					min = 4, max = 20, step = 1,
					order = 150,
					arg = {"boss3Frame",'size'},
				},
				["Font"] = {
					name = "Font",
					type = 'select',
					dialogControl = 'LSM30_Font',
					values = GetFontList,
					order = 200,
					arg = {"boss3Frame",'font'},
				},
				["FColor"] = {
					name = "Text Color",
					type = 'color',
					set = SetColorValue,
					get = GetColorValue,
					order = 350,
					arg = {"boss3Frame",'foreground'},
				},
				["TextAlpha"] = {
					name = "Text Alpha",
					type = 'range',
					min = 0, max = 1, step = .05,
					order = 400,
					arg = {"boss3Frame",'foregroundAlpha'},
				},
				["Outline"] = {
					name = "Outline",
					type = 'select',
					desc = "Set font outline.",
					values = {[''] = NONE, OUTLINE = VOICE_CHAT_NORMAL, THICKOUTLINE = "Thick"},
					order = 425,
					arg = {"boss3Frame",'outline'},
				},
				["BColor"] = {
					name = "Background",
					type = 'color',
					set = SetColorValue,
					get = GetColorValue,
					order = 450,
					arg = {"boss3Frame",'background'},
				},
				["BAlpha"] = {
					name = "Background Alpha",
					type = 'range',
					min = 0, max = 1, step = .05,
					order = 500,
					arg = {"boss3Frame",'backgroundAlpha'},
				},
				["Reset"] = {
					name = "Restore Defaults",
					func = RestoreDefaults,
					type = 'execute',
					order = 550,
					arg = "boss3Frame",
				},
			}
		},	
		
		
		["BOSS_FRAME4"] = {
			name = "Boss 4",
			type = 'group',
			order = 350,
			guiInline = true,
			set = SetValue,
			get = GetValue,
			args = {
				["Show"] = {
					name = "Show",
					type = 'toggle',
					order = 100,
					arg = {"boss4Frame",'show'},
					disabled = false,
				},
				["Font Size"] = {
					name = "Font Size",
					type = 'range',
					min = 4, max = 20, step = 1,
					order = 150,
					arg = {"boss4Frame",'size'},
				},
				["Font"] = {
					name = "Font",
					type = 'select',
					dialogControl = 'LSM30_Font',
					values = GetFontList,
					order = 200,
					arg = {"boss4Frame",'font'},
				},
				["FColor"] = {
					name = "Text Color",
					type = 'color',
					set = SetColorValue,
					get = GetColorValue,
					order = 350,
					arg = {"boss4Frame",'foreground'},
				},
				["TextAlpha"] = {
					name = "Text Alpha",
					type = 'range',
					min = 0, max = 1, step = .05,
					order = 400,
					arg = {"boss4Frame",'foregroundAlpha'},
				},
				["Outline"] = {
					name = "Outline",
					type = 'select',
					desc = "Set font outline.",
					values = {[''] = NONE, OUTLINE = VOICE_CHAT_NORMAL, THICKOUTLINE = "Thick"},
					order = 425,
					arg = {"boss4Frame",'outline'},
				},
				["BColor"] = {
					name = "Background",
					type = 'color',
					set = SetColorValue,
					get = GetColorValue,
					order = 450,
					arg = {"boss4Frame",'background'},
				},
				["BAlpha"] = {
					name = "Background Alpha",
					type = 'range',
					min = 0, max = 1, step = .05,
					order = 500,
					arg = {"boss4Frame",'backgroundAlpha'},
				},
				["Reset"] = {
					name = "Restore Defaults",
					func = RestoreDefaults,
					type = 'execute',
					order = 550,
					arg = "boss4Frame",
				},
			},
		}	
	},
}

function DeathClock:RegisterOptions()
	local global = LibStub('AceDBOptions-3.0'):GetOptionsTable(self.db)
	local registry = LibStub('AceConfigRegistry-3.0')
	local dialog = LibStub('AceConfigDialog-3.0')
	LibStub('AceConfig-3.0'):RegisterOptionsTable('Death Clock', options, {'Death Clock', 'Death Clock'})
	
	registry:RegisterOptionsTable('Death Clock Options', options)
	local main = dialog:AddToBlizOptions('Death Clock Options', 'Death Clock')
	local dataobj = LibStub:GetLibrary('LibDataBroker-1.1'):NewDataObject('Death Clock', {
		type = 'data source',
		text = 'Death Clock',
		label = 'Death Clock',
		icon = 'Interface\\Icons\\Spell_Holy_BorrowedTime',
	})
	
	if dataobj then
		dataobj.OnClick = function() InterfaceOptionsFrame_OpenToCategory(main) end
		dataobj.OnTooltipShow = function(tooltip)
			tooltip:AddLine('Death Clock')
			tooltip:AddLine('Display an estimated kill time for your target, focus target and any mobs in the blizzard boss frames.')
		end
	end
end

function DeathClock:SlashCommand(info)
	if info == "debug" then
		self:Debug()
	elseif info == "reset all" then
		RestoreFrameDefaults("targetFrame")
		RestoreFrameDefaults("focusFrame")
		RestoreFrameDefaults("boss1Frame")
		RestoreFrameDefaults("boss2Frame")
		RestoreFrameDefaults("boss3Frame")
		RestoreFrameDefaults("boss4Frame")
	elseif info == "reset target" then
		RestoreFrameDefaults("targetFrame")
	elseif info == "reset focus" then
		RestoreFrameDefaults("focusFrame")
	elseif info == "reset boss1" then
		RestoreFrameDefaults("boss1Frame")
	elseif info == "reset boss2" then
		RestoreFrameDefaults("boss2Frame")
	elseif info == "reset boss3" then
		RestoreFrameDefaults("boss3Frame")
	elseif info == "reset boss4" then
		RestoreFrameDefaults("boss4Frame")
	else
		print("Command List")
		print("     |cff3399FFreset all|r - Restores the default settings for all frames.")
		print("     |cff3399FFreset target|r - Restores the default settings for the target frame.")
		print("     |cff3399FFreset focus|r - Restores the default settings for the focus target frame.")
		print("     |cff3399FFreset boss1|r - Restores the default settings for the first boss frame.")
		print("     |cff3399FFreset boss2|r - Restores the default settings for the second boss frame.")
		print("     |cff3399FFreset boss3|r - Restores the default settings for the third boss frame.")
		print("     |cff3399FFreset boss4|r - Restores the default settings for the fourth boss frame.")
	end
end